#
# Files in this directory are automatically added to the devstack
# local.conf file, between a specific set of tags.
#
# Filenames must end with '.rc' to be recognized; sample.rc is
# ignored.
#
# A '\' is required in front of any devstack variables since all
# .rc files are parsed first (using eval).
#
# Meta section headings must be included in each file, such as:
# [[local|localrc]]
# as the order of inserting the files is not guaranteed.
#
# All files are inherently included by default - to exclude a file,
# add a variable 'FILENAME_IN_UPPERCASE_MINUS_RC=false' in trovestack.rc
# For Example: USING_VAGRANT=false (for the using_vagrant.rc file).
#
# Symbolic links are followed, so additional files can be loaded
# by placing them in an external directory and linking it in
# local.conf.d (this should allow complete flexibility in setting
# up testing options).
# For Example:
#   cd /path/to/trove/integration/scripts/local.conf.d
#   ln -s $HOME/local.conf.d local.conf.d
#   cp /path/to/my_conf.rc $HOME/local.conf.d


[[local|localrc]]
# Put regular devstack variables under this meta section heading.
# This section is written out to a file and sourced by devstack,
# so it can contain logic as well.

# The following section types should only contain ini file style
# section headings and name=value pairs
[[post-config|\$TROVE_CONF]]

[[post-config|\$TROVE_TASKMANAGER_CONF]]

[[post-config|\$TROVE_CONDUCTOR_CONF]]

[[post-config|\$TROVE_API_PASTE_INI]]
