[[post-config|\$TROVE_CONF]]
[profiler]
enabled = $ENABLE_PROFILER
trace_sqlalchemy = $PROFILER_TRACE_SQL

[[post-config|\$TROVE_TASKMANAGER_CONF]]
[profiler]
enabled = $ENABLE_PROFILER
trace_sqlalchemy = $PROFILER_TRACE_SQL

[[post-config|\$TROVE_CONDUCTOR_CONF]]
[profiler]
enabled = $ENABLE_PROFILER
trace_sqlalchemy = $PROFILER_TRACE_SQL

[[post-config|\$TROVE_GUESTAGENT_CONF]]
[profiler]
enabled = $ENABLE_PROFILER
trace_sqlalchemy = $PROFILER_TRACE_SQL

[[post-config|\$TROVE_API_PASTE_INI]]
[filter:osprofiler]
enabled = $ENABLE_PROFILER
hmac_keys = $PROFILER_HMAC_KEYS
