$TROVE_PRESENT_TAG
# Set some arguments for devstack.
#
# Note: This file contains autogenerated parts.
#       All lines are removed from between the tag/end of tag
#       markers (lines with '$MARKER_TOKEN' at beginning and end) and
#       are replaced by trovestack.
#       Edits to these sections will not persist.
#
#       See the '$USER_OPTS_TAG' section
#       for ways to insert user args into this file.
#

#
# This section is for things that belong in localrc
# It comes from $DEFAULT_LOCALRC
#
[[local|localrc]]

$LOCALRC_OPTS_TAG
$LOCALRC_OPTS_TAG_END

#
# User options here were inserted from the file USER_LOCAL_CONF
# (defaults to $USERHOME/.$LOCAL_CONF)
#

$USER_OPTS_TAG
$USER_OPTS_TAG_END

#
# Additional options here were inserted by trovestack
# automatically from files in $LOCAL_CONF_D
#

$ADD_OPTS_TAG
$ADD_OPTS_TAG_END
